# Generated by Django 4.1.9 on 2023-06-08 09:58

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("geography", "0001_initial"),
        ("wbhuman_resources", "0016_alter_employeehumanresource_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="absencerequest",
            name="crossborder_country",
            field=models.ForeignKey(
                blank=True,
                help_text="The country where this absence request will be held.",
                limit_choices_to={"level": 1},
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="absence_requests",
                to="geography.geography",
                verbose_name="Crossborder Country",
            ),
        ),
        migrations.AddField(
            model_name="absencerequesttype",
            name="crossborder_countries",
            field=models.ManyToManyField(
                blank=True,
                help_text="List of countries where crossborder activity is allowed",
                limit_choices_to={"level": 1},
                to="geography.geography",
                verbose_name="Countries",
            ),
        ),
        migrations.AddField(
            model_name="absencerequesttype",
            name="extra_notify_groups",
            field=models.ManyToManyField(
                blank=True,
                related_name="notified_absence_request_types",
                to="auth.group",
                verbose_name="Extra Notify Groups",
            ),
        ),
        migrations.AddField(
            model_name="absencerequesttype",
            name="is_country_necessary",
            field=models.BooleanField(default=False, verbose_name="Is country necessary"),
        ),
    ]
